﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Knapsack {
  /// <summary>
  /// An operator that relinks paths between knapsack solutions using a multiple guiding strategy.
  /// </summary>
  [Item("KnapsackMultipleGuidesPathRelinker", "An operator that relinks paths between knapsack solutions using a multiple guiding strategy.")]
  [StorableClass]
  public sealed class KnapsackMultipleGuidesPathRelinker : PathRelinker {
    [StorableConstructor]
    private KnapsackMultipleGuidesPathRelinker(bool deserializing) : base(deserializing) { }
    private KnapsackMultipleGuidesPathRelinker(KnapsackMultipleGuidesPathRelinker original, Cloner cloner) : base(original, cloner) { }
    public KnapsackMultipleGuidesPathRelinker() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new KnapsackMultipleGuidesPathRelinker(this, cloner);
    }

    public static ItemArray<IItem> Apply(IItem initiator, IItem[] guides, PercentValue n) {
      if (!(initiator is BinaryVector) || guides.Any(x => !(x is BinaryVector)))
        throw new ArgumentException("Cannot relink path because some of the provided solutions have the wrong type.");
      if (n.Value <= 0.0)
        throw new ArgumentException("RelinkingAccuracy must be greater than 0.");

      BinaryVector v1 = initiator.Clone() as BinaryVector;
      BinaryVector[] targets = new BinaryVector[guides.Length];
      Array.Copy(guides, targets, guides.Length);

      if (targets.Any(x => x.Length != v1.Length))
        throw new ArgumentException("At least one solution is of different length.");

      IList<BinaryVector> solutions = new List<BinaryVector>();
      for (int i = 0; i < v1.Length; i++) {
        // TODO: path relinking
      }

      IList<IItem> selection = new List<IItem>();
      if (solutions.Count > 0) {
        int noSol = (int)(solutions.Count * n.Value);
        if (noSol <= 0) noSol++;
        double stepSize = (double)solutions.Count / (double)noSol;
        for (int i = 0; i < noSol; i++)
          selection.Add(solutions.ElementAt((int)((i + 1) * stepSize - stepSize * 0.5)));
      }

      return new ItemArray<IItem>(selection);
    }

    protected override ItemArray<IItem> Relink(ItemArray<IItem> parents, PercentValue n) {
      if (parents.Length < 2)
        throw new ArgumentException("The number of parents is smaller than 2.");
      return Apply(parents[0], parents.Skip(1).ToArray(), n);
    }
  }
}
